'''记录备份'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os
import time


def StampFile(path):
    root, file = os.path.split(path)
    name, ext = os.path.splitext(file)
    root2 = os.path.join(root, '.pybak')
    os.makedirs(root2, exist_ok=True)

    tag = time.strftime('@%Y%m%d_%H%M%S')
    file2 = os.path.join(root2, name + tag + ext)
    return file2


class SaveArchive:
    def __init__(self, parent):
        self.io = parent.io
        self.get_saved = parent.get_saved
        parent.before_save.append(self.Saving)

    def Saving(self):
        if self.io.filename and not self.get_saved():
            self.io.writefile(StampFile(self.io.filename))



'''文件操作'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os
from os.path import dirname, basename
import subprocess
import tkinter as tk
from tkinter.simpledialog import askstring


class FileManager(tk.Menu):
    def __init__(self, parent):
        self.text = parent.text
        self.io = parent.io
        root = parent.root

        tk.Menu.__init__(self, root, tearoff=0)

        path = lambda: self.io.filename or os.getcwd() + os.sep # `io.filename` get `None` when in new file
        copy = lambda s: (root.clipboard_clear(), root.clipboard_append(s))
        insert = lambda s: self.text.insert('insert', s)

        self.add_command(label='Copy Fullname', command=lambda: copy(path()))
        self.add_command(label='Copy Filename', command=lambda: copy(basename(path())))
        self.add_command(label='Copy Dirname',  command=lambda: copy(dirname(path())))
        self.add_separator()
        self.add_command(label='Insert Fullname', command=lambda: insert("r'%s'" % path()))
        self.add_command(label='Insert Filename', command=lambda: insert("'%s'"  % basename(path())))
        self.add_command(label='Insert Dirname',  command=lambda: insert("r'%s'" % dirname(path())))
        self.add_separator()
        self.add_command(label='Open in Explorer', command=lambda: subprocess.Popen('explorer /select, "%s"' % path().replace('/', '\\')))
        self.add_command(label='Open in CMD',      command=lambda: subprocess.Popen('cmd /s /k pushd "%s"' % dirname(path())))
        # self.add_command(label='Run in CMD',       command=lambda: os.system('start "%s"'%io.filename)) # TODO
        self.add_separator()
        self.add_command(label='Rename File', command=self.Rename)
        self.add_command(label='Reload File', command=self.Reload)

        self.text.bind('<F2>', self.Rename)

        parent.amenu.insert_cascade(3, label='File Manager', menu=self)

    def Rename(self):
        basename2 = basename(self.io.filename)
        filename2 = askstring('Rename', 'Input new filename:', initialvalue=basename2, parent=self.text) # 允许输入相对路径
        if filename2:
            if filename2[-3:].lower() != '.py':
                filename2 += '.py'
            dirname2 = dirname(filename2)
            if dirname2: # 只输入文件名的时候路径为 ''
                os.makedirs(dirname2, exist_ok=True)
            os.rename(self.io.filename, filename2) # TODO 切换目录时工作路径未变化
            self.io.set_filename(os.path.abspath(filename2)) # TODO save状态变化

    def Reload(self): # TODO 重载记忆位置
        self.io.loadfile(self.io.filename)  # TODO rename也可以用这个方法



"""自动保存Untitled"""


import os
import time

if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


PATH = '.pybak/autosave.py'


class SaveUntitled:
    def __init__(self, parent):
        if hasattr(parent, 'write'): # is shell?
            return

        self.text = parent.text
        self.io = parent.io
        self.get_saved = parent.get_saved  # function

        parent.before_close.append(self.Backup)
        self.Reload()

    def Backup(self):
        if self.io.filename is None: # is untitled script?
            s = self.text.get('1.0', 'end-1c').rstrip()
            if s:
                os.makedirs('.pybak', exist_ok=True)
                with open(PATH, 'a') as f:
                    f.write(time.strftime('# autosave @ %Y-%m-%d %H:%M:%S\n') + s + '\n\n')
            elif os.path.isfile(PATH):
                os.remove(PATH)

    def Reload(self):
        if self.io.filename is None: # is untitled script?
            if os.path.isfile(PATH):
                with open(PATH) as f:
                    s = f.read()
                self.text.insert('1.0', s)
                self.text.mark_set('insert', '1.0')
                os.remove(PATH)



"""Shell换行"""


if __name__ == '__main__':
    import __init__
    __init__.test_shell(__file__)


class WrapShell:
    def __init__(self, parent):
        if hasattr(parent, 'write'): # is shell?
            self.text = parent.per.text  # TODO 研究一下作用原理

            parent.rmenu.add_separator()
            parent.rmenu.add_command(label='Wrap shell', command=self.ChangeWrap)
            parent.add_adv_menu('Wrap shell', self.ChangeWrap)

    def ChangeWrap(self):
        if self.text['wrap'] == 'none':
            self.text['wrap'] = 'char'
        else:
            self.text['wrap'] = 'none'



'''拖拽打开'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


class DragOpen:
    def __init__(self, parent):
        self.flist = parent.flist
        try:
            import windnd
            windnd.hook_dropfiles(parent.text, func=self.DragOpen)
        except Exception as e:
            print('import windnd error:', e)

    def DragOpen(self, files): # TODO 恢复记忆位置
        for file_b in files:
            file = file_b.decode('gbk')
            if file.endswith('.py'): # TODO 非py文件是否打开
                edit = self.flist.open(file)
                # TODO 由于滚动条存在导致有时候拖拽加载会闪退，增加下面两行可以避免
                edit.text.tag_add('sel', 'insert', 'insert+1c')
                edit.text.tag_remove('sel', '1.0', 'end')

                # TODO 是否可以
                # self.io.open(editFile=file)



'''快速正反搜索'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


# TODO 所有的搜搜窗口都不热点捕捉但是置前（不仅仅是F3启动）

import sys
if sys.version_info > (3, 6):
    from idlelib.search import _setup
else:
    from idlelib.SearchDialog import _setup


class QuickSearch:
    def __init__(self, parent):
        self.text = parent.text
        self.text.bind('<F3>',       lambda e: self.FindPrev(False))
        self.text.bind('<Shift-F3>', lambda e: self.FindPrev(True))

    def FindPrev(self, back):
        searchdialog = _setup(self.text)
        searchdialog.engine.backvar.set(back)
        searchdialog.find_again(self.text)
        searchdialog.top.grab_release() # 取消grab_set()窗口置顶



'''智能选取'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import re

sp = lambda c: eval(c.replace('.',',')) # Good!
jn = lambda x,y: '%i.%i'%(x,y) # Good!
lc = lambda s: jn(s.count('\n')+1, len(s)-s.rfind('\n')-1) # Good!

is_code = lambda s: s.split('#')[0].strip()


def FixTextSelect(root):
    tk = root.tk
    tk.eval('bind Text <Double-1> {catch {%W mark set insert sel.first}}') # See: "~\tcl\tk8.6\text.tcl"


def Select(text, c1, c2):
    text.tag_remove('sel', '1.0', 'end')
    text.tag_add('sel', c1, c2)
    text.mark_set('insert', c1)


def FindParen(s, c1='(', c2=')'):
    n1 = n2 = 0
    for n, c in enumerate(s):
        n1 += c in c1
        n2 += c in c2
        if n1 and n1 == n2:
            return n


def MatchSpan(r, s, n):
    for m in re.finditer(r, s):
        if m.start() <= n <= m.end():
            return m.span()


def LineType(line):
    indent = re.match(r' *', line).end()
    code = line[indent:]
    if code.startswith('#'):
        return indent, 'comment'
    elif code:
        return indent, 'code'
    else:
        return indent, 'empty'


def Selecting(e):
    text = e.widget
    text.tag_remove('hit', '1.0', 'end')

    if 'STRING' in text.tag_names('current'):
        st, ed = text.tag_prevrange('STRING', 'current+1c') # See: `idlelib.squeezer.Squeezer.squeeze_current_text_event`
        ed2 = text.index(ed + '-1c')
        cur = text.index('current')
        if cur in (st, ed2):
            return Select(text, st, ed)

    cur = text.index('current') # 当用insert时光标位置为自动选区的最开始
    col = sp(cur)[1]
    line = text.get('current linestart', 'current lineend')

    c = text.get('current') # charset: !"#$%&'()*+,-./:;<=>?@[\]^_`{|}~
    if c == ':':
        if is_code(line[col+1:]):
            Select(text, 'current', 'current+1c')
        else:
            indent = re.match(r' *', line).end()
            ss = text.get('current linestart+1l', 'end').split('\n')
            n = 1
            for r, row in enumerate(ss):
                indent2, typ = LineType(row)
                if typ == 'code' and indent2 <= indent:
                    break
                if typ != 'empty' and indent2 > indent:
                    n = r + 2
            Select(text, 'current linestart', 'current linestart+%dl' % n)

    elif c == ' ':
        indent = re.match(r' *', line).end()
        if col < indent:
            ss = text.get('current linestart+1l', 'end').split('\n')
            for n1, row in enumerate(ss):
                indent2, typ = LineType(row)
                if typ == 'empty':
                    break
            ss = text.get('1.0', 'current linestart-1c').split('\n')
            for n0, row in enumerate(reversed(ss)):
                indent2, typ = LineType(row)
                if typ == 'empty':
                    break
            Select(text, 'current linestart-%dl' % n0, 'current linestart+%dl' % (n1 + 1))
        else:
            p1, p2 = MatchSpan(r' +', line, col)
            Select(text, 'current linestart+%dc' % p1, 'current linestart+%dc' % p2)

    elif re.match(r'\w', c):
        p1, p2 = MatchSpan(r'\w+', line, col)
        Select(text, 'current linestart+%dc' % p1, 'current linestart+%dc' % p2)
        word = line[p1:p2]
        s = text.get('1.0', 'end')
        for m in re.finditer(r'\b%s\b' % word, s):
            p1, p2 = m.span()
            text.tag_add('hit', lc(s[:p1]), lc(s[:p2])) # 如果使用`1.0+nc`字符偏移会命中Squeezer导致错位

    elif c == '\n':
        Select(text, 'current linestart', 'current+1c')

    elif c == '#': # for # test # text #
        if 'COMMENT' in text.tag_names('current') and \
           text.index('current') == text.tag_prevrange('COMMENT', 'current+1c')[0]:
            p1, p2 = MatchSpan(r' *#', line, col)
            if p1 > 0:
                Select(text, 'current linestart+%dc' % p1, 'current lineend')
            else:
                Select(text, 'current linestart', 'current linestart+1l')
        else:
            Select(text, 'current', 'current+1c')

    elif c in '\'"`': # quote in comment
        s = text.get('current+1c', 'end')
        n = s.find(c)
        Select(text, 'current', 'current+%dc' % (n + 2))

    elif c in '([{':
        # idle内自建的方法太复杂 idlelib.hyperparser.HyperParser.get_surrounding_brackets
        c1 = c
        c2 = ')]}'['([{'.index(c1)]
        s = text.get('current', 'end')
        n = FindParen(s, c1, c2)
        Select(text, 'current', 'current+%dc' % (n + 1))

    elif c in ')]}':
        c1 = c
        c2 = '([{'[')]}'.index(c1)]
        s = text.get('1.0', 'current+1c')
        n = FindParen(reversed(s), c1, c2)
        Select(text, 'current-%dc' % n, 'current+1c')

    elif c == '\\': # "\ooo" "\xhh" "\uxxxx" "\Uxxxxxxxx"
        c2 = text.get('current+1c')
        n = {'x': 4, 'u': 6, 'U': 10}.get(c2, 2)
        Select(text, 'current', 'current+%dc' % n)

    elif c == '.':
        s = text.get('current+1c', 'current lineend')
        n = re.match('\s*\w*', s).end()
        Select(text, 'current', 'current+%dc' % (n + 1))

    elif c == ',': # for test: (a, b(c, d=5), e); for a, b in c; a, b = c, 'd'
        s = text.get('current+1c', 'end')
        n1 = n2 = 0
        for n, c1 in enumerate(s):
            if n1 == n2 and c1 in ',)]}':
                break
            n1 += c1 in '([{'
            n2 += c1 in ')]}'
        Select(text, 'current', 'current+%dc' % (n + 1))

    else: # charset: !$%&*+,-/;<=>?@^|~  for test: s.split('\n')[:ln]
        p1, p2 = MatchSpan(r'[^\w()[\]{}\'" ]+', line, col) # 不"粘"括号空格和引号
        Select(text, 'current linestart+%dc' % p1, 'current linestart+%dc' % p2)
        # Select(text, 'current', 'current+1c')


class SmartSelect:
    def __init__(self, parent):
        FixTextSelect(parent.root)
        parent.text.bind('<Double-Button-1>', Selecting)



"""显示版本信息和自动升级"""


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os

import sys
if sys.version_info > (3, 6):
    from idlelib.textview import view_text
else:
    from idlelib.textView import view_text


def read_relative(filename):
    root = os.path.dirname(__file__)
    p1 = os.path.join(root, 'docs', filename)
    p2 = os.path.join(os.path.dirname(root), filename)
    path = p1 if os.path.isfile(p1) else p2
    with open(path, encoding='u8') as f:
        return f.read()


class About:
    def __init__(self, parent):
        self.text = parent.text
        parent.add_adv_menu('About', self.about)

    def about(self):
        view_text(self.text, 'idlea v' + read_relative('version.txt'), read_relative('readme.md'))

    def check(self): # TODO 检查版本是否最新
        pass

    def upgrade(self): # TODO 自动升级/自动获取当前Python的pip
        pass



'''自动备份'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os

TICK = 7000


class AutoSave:
    def __init__(self, parent):
        if hasattr(parent, 'write'): # is shell?
            return

        self.text = parent.text
        self.root = parent.root
        self.io = parent.io
        self.get_saved = parent.get_saved # function

        self.fixfixlastline()

        self.root.after(TICK, self.Backup)

        self.AfterOpen()
        parent.after_save.append(self.AfterSave)
        parent.before_close.append(self.BeforeClose)

    def fixfixlastline(self): # append '\n' at last line **after** cursor
        def fixlastline2():
            self.text.mark_gravity('insert', 'left')
            fixlastline()
            self.text.mark_gravity('insert', 'right')
        fixlastline = self.io.fixlastline
        self.io.fixlastline = fixlastline2

    def Backup(self, loop=True):
        filename = self.io.filename
        if not self.io.text: # will set as None in io.close()
            return # break loop only editor closed
        if filename:
            filename_bak = os.path.splitext(filename)[0] + '.pybak'
            if self.get_saved():
                if os.path.isfile(filename_bak):
                    os.remove(filename_bak)
            else:
                self.io.writefile(filename_bak)
        if loop: # don't stop loop while no filename, new file maybe saved later
            self.root.after(TICK, self.Backup)

    def AfterSave(self):
        self.Backup(False)

    def BeforeClose(self):
        self.Backup(False)

    def AfterOpen(self): # TODO 恢复存档
        filename = self.io.filename
        if filename:
            filename_bak = os.path.splitext(filename)[0] + '.pybak'
            if os.path.isfile(filename_bak):
                self.FakeLoadFile(filename_bak)

    def FakeLoadFile(self, filename_bak): # is that too.. dirty?
        fake = lambda *args: None
        temp = self.io.set_filename, self.io.reset_undo, self.io.updaterecentfileslist
        self.io.set_filename = self.io.reset_undo = self.io.updaterecentfileslist = fake

        self.text.undo_block_start()
        self.text.insert('1.0', '') # In order to undo, the cursor can move to the top.
        self.io.loadfile(filename_bak)
        self.text.undo_block_stop()

        self.io.set_filename, self.io.reset_undo, self.io.updaterecentfileslist = temp



"""重复选中文本/删除当前行"""

if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


class DuplicateLines:
    def __init__(self, parent):
        self.text = parent.text
        self.text.bind('<Control-d>', self.copy)
        self.text.bind('<Control-D>', self.delete)

    def copy(self, evt):
        if self.text.tag_ranges('sel'):
            self.text.insert('sel.last', self.text.get('sel.first', 'sel.last'))
        else:
            self.text.insert('insert+1l linestart', self.text.get('insert linestart', 'insert+1l linestart'))
        return 'break'

    def delete(self, evt):
        if self.text.tag_ranges('sel'):
            self.text.delete('sel.first', 'sel.last')
        else:
            ins = self.text.index('insert')
            self.text.delete('insert linestart', 'insert+1l linestart')
            self.text.mark_set('insert', ins)
        return 'break'



'''位置记录'''

# TODO 未保存文档但更新位置，可能导致光标位置指向错误？


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os
import csv

import sys
if sys.version_info > (3, 6):
    from idlelib.config import idleConf
else:
    from idlelib.configHandler import idleConf

userdir = idleConf.GetUserCfgDir()
rc_path = os.path.join(userdir, 'recent-saved.lst')
if not os.path.exists(rc_path):
    open(rc_path, 'w').close()


def ReadData():
    with open(rc_path, encoding='u8', newline='') as f:
        data = list(csv.reader(f))
    return data


def SaveData(data):
    with open(rc_path, 'w', encoding='u8', newline='') as f:
        writer = csv.writer(f)
        writer.writerows(data)


def FindInData(data, file):
    for n, (cur, file2) in enumerate(data):
        if file == file2:
            return n, cur.replace(':', '.')


def Update(file, cur, save=True):
    cur = cur.replace('.', ':')
    data = ReadData()
    ret = FindInData(data, file)
    if ret:
        n = ret[0]
        data[n][0] = cur
        if save:
            data.insert(0, data.pop(n)) # move to first
    else:
        data.insert(0, [cur, file])
    SaveData(data)


class RecentSaved:
    def __init__(self, parent):
        self.text = parent.text
        self.io = parent.io
        self.AfterOpen()
        parent.after_save.append(self.OnSave)
        parent.before_close.append(self.OnClose)

    def AfterOpen(self):
        if not self.io.filename:
            return
        data = ReadData()
        ret = FindInData(data, self.io.filename)
        if ret:
            cur = ret[1]
            text = self.text
            text.mark_set('insert', cur)
            text.see('insert linestart') # See line start to prevent XScrollBar at right.
            text.tag_remove('sel', '1.0', 'end')
            text.tag_add('sel', 'insert linestart', 'insert linestart+1l')
        else:
            data.insert(0, ['1:0', self.io.filename])
            SaveData(data)

    def OnSave(self, save=True):
        if not self.io.filename:
            return
        cur = self.text.index('insert')
        Update(self.io.filename, cur, save)

    def OnClose(self):
        self.OnSave(False)



"""纵向选择"""


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


from tkinter.simpledialog import askstring

# See: "~\tcl\tk8.6\text.tcl"
tcl1 = 'bind Text <B1-Motion> {set tk::Priv(x) %x; set tk::Priv(y) %y; tk::TextSelectTo %W %x %y}'
tcl2 = 'bind Text <B1-Motion> {tk::TextButton1 %W %x %y; %W tag remove sel 0.0 end}'

sp = lambda c: eval(c.replace('.',',')) # Good!
jn = lambda x,y: '%i.%i'%(x,y) # Good!
sort = lambda p1, p2: ((min(p1[0], p2[0]), min(p1[1], p2[1])), (max(p1[0], p2[0]), max(p1[1], p2[1])))


def enable(func):
    def wrapper(self, *args, **kwargs):
        if self.enable:
            return func(self, *args, **kwargs)
    return wrapper


class SelectVertical:
    def __init__(self, parent):
        self.text = parent.text
        self.root = parent.root

        self.enable = False

        self.text.bind('<Button-1>',        self.down)
        self.text.bind('<ButtonRelease-1>', self.up, add=True)
        self.text.bind('<B1-Motion>',       self.move)

        # self.text.bind('<KeyPress-Alt_L>',   self.setup)
        # self.text.bind('<KeyRelease-Alt_L>', self.cancel)

        parent.rmenu.add_separator()
        parent.rmenu.add_command(label='Select Vertical', command=self.setup)
        parent.add_adv_menu('Select Vertical', self.setup)

    def index(self, tag):
        return eval(self.text.index(tag).replace('.', ','))

    def setup(self, evt=-1):
        self.root.eval(tcl2)
        self.enable = True

    def cancel(self, evt=-1):
        self.root.eval(tcl1)
        self.enable = False

    def _down(self):
        self.p1 = self.index('insert')

    @enable
    def down(self, evt):
        self.text.after_idle(self._down)

    @enable
    def up(self, evt):
        self.p2 = self.index('insert')
        self.text.after_idle(self.select, self.p1, self.p2)
        if self.p1[0] != self.p2[0]: # multi-line selected
            s = askstring('Change to?', 'Input text:', parent=self.root)
            self.replace(self.p1, self.p2, s)
            self.text.focus() # grab focus after askstring dialog
            self.cancel()

    @enable
    def move(self, evt):
        p2 = self.index('insert')
        self.text.after_idle(self.select, self.p1, p2)

    def span(self, p1, p2):
        p1, p2 = sort(p1, p2)
        for r in range(p1[0], p2[0] + 1):
            yield jn(r, p1[1]), jn(r, p2[1])

    def select(self, p1, p2):
        for c1, c2 in self.span(p1, p2):
            self.text.tag_add('sel', c1, c2)

    def replace(self, p1, p2, new):
        text = self.text
        if new is None: # click cancel
            s = [text.get(c1, c2) for c1, c2 in self.span(p1, p2)]
            text.clipboard_clear()
            text.clipboard_append('\n'.join(s))
        else:
            text.undo_block_start()
            [text.delete(c1, c2) for c1, c2 in self.span(p1, p2)]
            text.undo_block_stop()
        if new: # do copy
            text.undo_block_start()
            [text.insert(c1, new) for c1, c2 in self.span(p1, p2)]
            text.undo_block_stop()



'''交换复制'''

# TODO 改为Ctrl-V的时候交换
# TODO Shell中没有标色


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


TAG = 'BREAK'

class SwapCopy:
    def __init__(self, parent):
        self.text = parent.text
        self.range = None
        self.text.bind('<Control-Shift-X>', self.Swaping)

    def Swaping(self, e):
        text = self.text
        range = text.tag_ranges('sel')
        if range:
            if self.range:
                (c1, c2), (c3, c4) = self.range, range
                if text.compare(c1, '>', c3): # 先替换前面的会导致后面的坐标错误
                    (c1, c2), (c3, c4) = (c3, c4), (c1, c2)
                if text.compare(c2, '>', c3):
                    self.Clear()
                    return
                s1 = text.get(c1, c2)
                s2 = text.get(c3, c4)
                text.undo_block_start()
                text.delete(c3, c4)
                text.insert(c3, s1)
                text.delete(c1, c2)
                text.insert(c1, s2)
                text.undo_block_stop()
                self.Clear()
            else:
                self.range = range # first
                text.tag_add(TAG, *range)
        else:
            self.Clear()

    def Clear(self):
        self.range = None  # clear setup
        self.text.tag_remove(TAG, '1.0', 'end')



"""大小写转换"""


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


class UpperCase:
    def __init__(self, parent):
        self.text = parent.text
        self.text.bind('<Control-u>', self.callback)

    def callback(self, evt):
        text = self.text
        if text.tag_ranges('sel'):
            old = text.get('sel.first', 'sel.last')
            st = text.index('sel.first')
            ed = text.index('sel.last')
            s1 = old.lower()
            s2 = old.upper()
            if s1 != old:
                self.replace(st, ed, s1)
            elif s2 != old:
                self.replace(st, ed, s2)

    def replace(self, st, ed, s):
        text = self.text
        text.undo_block_start()
        text.delete(st, ed)
        text.insert(st, s)
        text.tag_add('sel', st, ed)
        text.mark_set('insert', st)
        text.undo_block_stop()



'''运行选中'''

if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import tkinter as tk

SRC_ST = '1.0'
SRC_ED = 'end'
SEL_ST = 'sel.first'
SEL_ED = 'sel.last'
INS_ST = 'insert'


class RunSelected(tk.Menu):
    def __init__(self, parent):
        if hasattr(parent, 'write'): # is shell?
            return

        tk.Menu.__init__(self, parent.menubar, tearoff=0)
        self.text = parent.text
        self.flist = parent.flist
        self.rmenu = parent.rmenu
        self.MakeMenu(4)
        self.Binding()
        parent.amenu.insert_cascade(3, label='Run Selected', menu=self) # TODO 单元测试时插入偏差

    def Run(self, st, ed):
        code = self.text.get(st, ed)

        if not code:
            self.text.tag_add('sel', 'insert linestart', 'insert+1l linestart')
            self.text.mark_set('insert', 'insert linestart')
            code = self.text.get(st, ed)

        if code.startswith(' '):
            code = 'if 1:\n' + code

        # ref: idlexlib.extensions.RunSelection
        msg = '# Run Region [%s-%s]\n' % (self.text.index(st), self.text.index(ed))
        shell = self.flist.open_shell()
        console = shell.interp.tkconsole
        console.text.insert('insert', msg.replace('.', ':'))
        shell.interp.runcode(code)
        # TODO 报错位置和真实行号对应

    def MakeMenu(self, pos):
        self.add_command(label='Run Selected',    command=lambda: self.Run(SEL_ST, SEL_ED))
        self.add_command(label='Run to Cursor',   command=lambda: self.Run(SRC_ST, INS_ST))
        self.add_command(label='Run from Cursor', command=lambda: self.Run(INS_ST, SRC_ED))
        self.add_command(label='Run this Script', command=lambda: self.Run(SRC_ST, SRC_ED))

        rmenu = self.rmenu # reverse order
        rmenu.insert_command(pos, label='Run this Script', command=lambda: self.Run(SRC_ST, SRC_ED))
        rmenu.insert_command(pos, label='Run from Cursor', command=lambda: self.Run(INS_ST, SRC_ED))
        rmenu.insert_command(pos, label='Run to Cursor',   command=lambda: self.Run(SRC_ST, INS_ST))
        rmenu.insert_command(pos, label='Run Selected',    command=lambda: self.Run(SEL_ST, SEL_ED))
        rmenu.insert_separator(pos)

    def Binding(self):
        text = self.text
        text.bind('<Shift-F5>', lambda e: self.Run(SEL_ST, SEL_ED) if text.tag_ranges('sel') else self.Run(SRC_ST, SRC_ED))
        text.bind('<Shift-F6>', lambda e: self.Run(SRC_ST, INS_ST))
        text.bind('<Shift-F7>', lambda e: self.Run(INS_ST, SRC_ED))



'''当选区为空时复制/剪切复制当前行'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


class CopyNoSelect:
    def __init__(self, parent):
        self.text = parent.text
        parent.before_copy.insert(0, self.FixNoSelect) # make sure run first

    def FixNoSelect(self):
        if not self.text.tag_ranges("sel"):
            self.text.tag_add('sel', 'insert linestart', 'insert+1l linestart')
            self.text.mark_set('insert', 'insert linestart')



'''快速打开选中的文本'''

import os

if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


class QuickOpen:
    def __init__(self, parent):
        self.text = parent.text
        self.top = parent.top

        parent.rmenu.add_separator()
        # parent.rmenu.add_command(label='Open', command=self.Open)
        parent.rmenu.add_command(label='Start', command=self.Start)

        # parent.add_adv_menu('Quick Open', self.Open)
        parent.add_adv_menu('Quick Start', self.Start)

        print(os.getcwd())

    # def Open(self):
    #     if not self.text.tag_ranges('sel'):
    #         return
    #     s = self.text.get('sel.first', 'sel.last')
    #     os.popen('explorer ' + s)

    def Start(self):
        if not self.text.tag_ranges('sel'):
            return
        s = self.text.get('sel.first', 'sel.last')
        os.popen('start ' + s)


'F:\lsx\coding\idle\idlealib\scripts'
'scripts\readme.md'
"F:\lsx\代码说明.docx"
"notepad 1.txt"
"C:\\"
"cmd /k ipconfig"

# todo 相对文件夹打不开



import imp
imp.load_package('pkg', '../__init__.py').run()

# another way
# import importlib
# spec = importlib._bootstrap_external.spec_from_file_location('pkg', '../__init__.py')
# pkg  = importlib._bootstrap._load(spec)
# pkg.run()

# another way
# import importlib.util
# spec = importlib.util.spec_from_file_location('pkg', '../__init__.py')
# pkg = importlib.util.module_from_spec(spec)
# spec.loader.exec_module(pkg)
# pkg.run()



'''将选中代码转化为二维码'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import tkinter as tk
from PIL import ImageTk

import qrcode

# TODO 过长文本保护（Unicode长度）/翻页显示
# TODO 调整弹出位置


def qrmake(qrstr): # max about 2340 chars
    qr = qrcode.QRCode(box_size=2, border=2)
    qr.add_data(qrstr)
    qr.best_fit()
    qr.makeImpl(False, 3) # 3 or 6
    img = qr.make_image()
    return img


class ViewWindow(tk.Toplevel):
    def __init__(self, parent, title, text):
        super().__init__(parent)
        self.transient(parent) # hide minimize/maximize icon
        x = parent.winfo_rootx() + 10
        y = parent.winfo_rooty() + 10
        self.geometry('+%d+%d'%(x, y))
        self.title(title)
        self.img = ImageTk.PhotoImage(image=qrmake(text), master=self)
        qrc = tk.Label(self, image=self.img)
        qrc.pack()

        self.focus()
        self.bind('<Return>', lambda e: self.destroy())
        self.bind('<Escape>', lambda e: self.destroy())


class ShareQRCode:
    def __init__(self, parent):
        self.text = parent.text
        self.top = parent.top

        self.text.bind('<<share-qrcode>>', self.Post)

        parent.rmenu.add_separator()
        parent.rmenu.add_command(label='Share Code', command=self.Post)

        parent.add_adv_menu('Share QRCode', self.Post, sp=True)

    def Post(self):
        s = self.text.get('sel.first', 'sel.last') or self.text.get('1.0', 'end')
        s = s[:2000] # TODO 过长保护提示
        brief = s if len(s) < 21 else s[:10] + ' ... ' + s[-10:]
        ViewWindow(self.top, '%d chars: %s'%(len(s), brief), s)



'''光标记录'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


class CursorHistory:
    def __init__(self, parent):
        self.text = parent.text
        self.pointer = 0
        self.history = ['1.0'] # TODO 第一次是1.0或上次关闭前记录的位置
        self.Binding()

        if parent.amenu.index('end') is not None:
            parent.amenu.insert_separator(0)
        parent.amenu.insert_command(0, label='Forward', command=lambda: self.Move(1))
        parent.amenu.insert_command(0, label='Back', command=lambda: self.Move(-1))

    def Add(self, e):
        # TODO 当文本变化时平移历史记录
        # TODO 记录文件名+位置
        # TODO 只记录行数
        cur = self.text.index('insert')
        if cur != self.history[self.pointer]:
            self.pointer += 1
            self.history = self.history[:self.pointer] + [cur]

    def Move(self, n):
        if -1 < self.pointer + n < len(self.history):
            self.pointer += n
            self.text.see(self.history[self.pointer])
            self.text.mark_set('insert', self.history[self.pointer])
        return 'break'

    def Binding(self):
        text = self.text
        text.bind('<ButtonRelease-1>', self.Add, add=True)
        text.bind('<Alt-Left>',  lambda e: self.Move(-1))
        text.bind('<Alt-Right>', lambda e: self.Move(1))



"""打印回车符换成换行符"""
# 解决进度条库在IDLE中全部显示在一行中的问题


if __name__ == '__main__':
    import __init__
    __init__.test_shell(__file__)


import re


CLEAR_LINE = 1


class PrintCr:
    def __init__(self, parent):
        if hasattr(parent, 'write'): # is shell?
            self.text = parent.per.bottom
            self._write, parent.write = parent.write, self.write

    def write(self, s, tags=()):
        s = s.replace('\r\n', '\n')
        if '\r' in s and CLEAR_LINE: # 清除行和替换为换行两种模式选择
            self.text.delete('iomark linestart', 'iomark lineend')
            s = re.sub(r'.*\r', '', s)
        self._write(s.replace('\r', '\n'), tags)


""" TEST CODE:
print('12345\rabc')
for i in range(1, 10):
    __import__('time').sleep(0.2)
    print(f'\rprocess: {i/10:.0%}', end='', file=__import__('sys').stderr)

"""



"""保存时删除后续空格"""


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import re


class TrimTrailing:
    def __init__(self, parent):
        self.text = parent.text
        parent.before_save.append(self.callback)

    def get_line(self, index):
        return int(self.text.index(index).split('.')[0])

    def callback(self):
        text = self.text

        ln_end = self.get_line('end-1c')
        ln_ins = self.get_line('insert')
        if text.tag_ranges('sel'):
            ln_ins = -1

        text.undo_block_start()

        for ln in range(1, ln_end + 1):
            # skip current line or in selection mode
            if ln != ln_ins:
                s = text.get('%d.0' % ln, '%d.end' % ln)
                m = re.search(' +$', s)
                if m:
                    text.delete('%d.%d' % (ln, m.start()), '%d.end' % ln)

        text.undo_block_stop()



'''历史剪切板'''

# TODO 不能在多个IDLE中共享


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import tkinter as tk


class RecentClipboard(tk.Menu):
    def __init__(self, parent):
        tk.Menu.__init__(self, parent.menubar, tearoff=0)
        self.text = parent.text
        self.root = parent.root

        self.data = []

        self.Binding()
        self.Add()

        parent.rmenu.insert_cascade(3, label='History', menu=self)
        parent.amenu.insert_cascade(3, label='Paste from History', menu=self)

        try:
            self.Add(parent.top.clipboard_get())
        except: # 当剪切板为图像时无法获取
            pass

        parent.before_copy.append(self.Add)

    def Add(self, s=''):
        s = s or self.text.get('sel.first', 'sel.last')

        if not s: # empty string
            return

        if s in self.data:
            self.data.remove(s)
        self.data.insert(0, s)

        self.delete(0, 'end')
        for i, s in enumerate(self.data): # TODO 限制个数
            label, callback = self.Prep(s)
            self.add_command(label=label, command=callback) # TODO 增加0-9A-Z快捷键

    def Prep(self, s):
        def paste():
            self.root.clipboard_clear()
            self.root.clipboard_append(s)
            self.text.event_generate('<<Paste>>')
            self.Add(s)

        s1 = s.replace('\n', '\\n')
        if len(s1) > 45:
            s1 = s1[:20] + ' ... ' + s1[-20:]
        s2 = 'len: %d, %s' % (len(s), s1)

        return s2, paste

    def Post(self, e):
        x0, y0 = self.text.winfo_rootx(), self.text.winfo_rooty()
        x, y, w, h = self.text.bbox('insert')
        self.post(x0 + x, y0 + y + h)

    def Binding(self):
        self.text.bind('<Control-Shift-V>', self.Post)



"""添加和取消注释"""


import os
import re


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


class AutoComment:
    def __init__(self, parent):
        self.text = parent.text
        self.get_region = parent.get_region
        self.set_region = parent.set_region
        self.text.bind('<Control-/>', self.callback)

    def callback(self, evt):
        head, tail, chars, lines = self.get_region()
        lines.pop()

        prefix = os.path.commonprefix(lines)
        m = re.match(r'(\s*)(##|# |#)?', prefix)
        header, comment = m.groups()
        if comment:
            start = len(header + comment)
        else:
            start = len(header)
            header += '# '
        lines = [header + line[start:] for line in lines]

        self.set_region(head, tail, chars, lines + [''])
        return 'break'



'''Shell多行运行'''


if __name__ == '__main__':
    import __init__
    __init__.test_shell(__file__)


# TODO 自动格式化代码之前的空格
# TODO 增加到菜单（开关）


class RunMultiLine:
    def __init__(self, parent):
        if hasattr(parent, 'write'): # is shell?
            self.parent = parent
            self.text = parent.text
            self.text.bind('<<Paste>>', self.AfterPaste, add=True)

    def AfterPaste(self, evt):
        self.text.after(10, self.Setup)

    def Setup(self):
        text = self.text
        self.codes = text.get('iomark', 'end-1c').split('\n')
        if len(self.codes) > 1 and text.index('insert') == text.index('end-1c'): # 多行复制且光标在最后
            text.delete('iomark', 'end-1c')
            text.after(10, self.Run)

    def Run(self): # See: idlelib.pyshell.PyShell.enter_callback
        p = self.parent
        text = self.text
        if self.codes and not p.canceled:
            while p.executing:
                text.after(10, self.Run)
                return
            code = self.codes.pop(0)
            text.mark_gravity('iomark', 'left') # without this it will run code skipped sometime. un-know reason.
            text.insert('insert', code)
            p.color.recolorize()
            if self.codes: # dont run last code.
                text.event_generate('<Return>')
                text.delete('end-1l', 'end-1c') # delete auto indent spaces.
            text.see('insert')
            text.after(10, self.Run)



'''横向滚动条'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import sys
PY37 = sys.version_info > (3, 7)

import tkinter as tk


class ScrollHorizontal:
    def __init__(self, parent):
        text = parent.text
        frame = parent.text_frame

        # TODO 会导致拖拽打开文件时闪退
        hbar = tk.Scrollbar(frame, orient='h')
        hbar['command'] = text.xview
        text['xscrollcommand'] = hbar.set
        if PY37:
            hbar.grid(row=2, column=1, sticky='nsew')
        else:
            hbar.pack(fill='x', side='bottom')
        # TODO 参考 idlelib.textview.ViewWindow 的 AutoHiddenScrollbar 方法



"""主程序入口"""

# Usage: python -m idlealib


def FixPath():
    # To fix `python -m idlealib` cant find `__init__`.
    import os
    import sys
    path = os.path.dirname(__file__)
    if path not in sys.path:
        sys.path.insert(0, path)


if __name__ == '__main__':
    FixPath()
    import __init__
    __init__.run()



"""IDLE-Advance: IDLE Useful Extensions

运行__main__.py获得一个加载所有插件的IDLE-Advance的示例文件。
运行__init__.py获得一个打开自身脚本并加载所有插件的editor的例子。
分别运行idlealib目录下的扩展文件，可以得到一个打开自身的editor或shell的例子。
如果需要停用部分扩展，将对应的脚本移出目录后重启IDLE即可。

See also: https://github.com/znsoooo/IDLE-Advance

MIT license. Copyright (c) 2021-2022 Lishixian (znsoooo). All Rights Reserved.

"""


import os
import sys

PY36 = sys.version_info > (3, 6)
# `abspath` for open in cmd like `python __init__.py` to open script.
EXTENSIONS = os.listdir(os.path.dirname(os.path.abspath(__file__)))

__author__  = 'Lishixian <lsx7@sina.com>'
__credits__ = 'See at: https://github.com/znsoooo/IDLE-Advance'
__date__    = '2022'
__version__ = '1.2'

__all__ = ['run', 'test_editor', 'test_shell', 'PY36']


# - Functions ----------------------------------------


def fix_path():
    # To fix open shell (call run() without sys.argv) or self.load_extension(name) will not work.
    import sys
    path = os.path.dirname(__file__)
    if path not in sys.path:
        sys.path.insert(0, path)


def wrap_function(func, before=(), after=()):
    def wrapper(*args, **kwargs):
        # print(func.__name__) # for test
        [f() for f in before]
        ret = func(*args, **kwargs)
        [f() for f in after]
        return ret
    return wrapper


# - idleConf ----------------------------------------


if not PY36:
    from functools import partial
    from idlelib.configHandler import idleConf

    idleConf.GetOption = partial(idleConf.GetOption, warn_on_default=False)

    # idleConf.userCfg['extensions'].set('ZzDummy', 'enable', 'true') # fix do not show many warning in py34 and py35
    #
    # if idleConf.userCfg['extensions'].has_option('ZzDummy', 'enable'):
    #     ret = idleConf.userCfg['extensions'].Get('ZzDummy', 'enable', type='bool')
    #     print(ret)
    #
    # print(idleConf.GetOption('extensions', 'ZzDummy', 'enable', default=True, type='bool'))
    # sys.exit(0)

    # idleConf._GetOption, idleConf.GetOption = idleConf.GetOption, print
    # def GetOption(configType, section, option, *args, **kw):
    #     if section.lower() == 'ZzDummy'.lower():
    #         print((configType, section, option))
    #         return False
    #     ret = idleConf._GetOption(configType, section, option, *args, **kw)
    #     return ret
    # idleConf.GetOption = GetOption


# - Calltip ----------------------------------------


if PY36:
    import idlelib.calltip
    from idlelib.calltip import Calltip
else:
    import idlelib.CallTips
    from idlelib.CallTips import CallTips as Calltip


# idlelib.calltip._MAX_LINES = 999
# idlelib.calltip._MAX_COLS = 999


class MyCalltip(Calltip):
    def __init__(self, editwin=None):
        super().__init__(editwin)
        editwin.ctip = self # make hook

    # def fetch_tip(self, expression):
    #     self.expression = expression
    #     self.argspec = super().fetch_tip(expression)
    #     return self.argspec


if PY36:
    idlelib.calltip.Calltip = MyCalltip
else:
    idlelib.CallTips.CallTips = MyCalltip


# - AutoComplete ----------------------------------------


if sys.version_info < (3, 8):
    if PY36:
        import idlelib.autocomplete
        from idlelib.autocomplete import AutoComplete
    else:
        import idlelib.AutoComplete
        from idlelib.AutoComplete import AutoComplete

    import keyword

    class MyAutoComplete(AutoComplete):
        def fetch_completions(self, what, mode):
            ret = super().fetch_completions(what, mode)
            ATTRS = int(sys.version_info < (3, 7))
            if mode == ATTRS and what == '':
                for lst in ret[:2]:
                    lst.extend(v for v in keyword.kwlist if v not in lst) # `None/True/False` are repetitive.
                    lst.sort()
            return ret

    if PY36:
        idlelib.autocomplete.AutoComplete = MyAutoComplete
    else:
        idlelib.AutoComplete.AutoComplete = MyAutoComplete


# - IOBinding ----------------------------------------


if PY36:
    import idlelib.iomenu
    from idlelib.iomenu import IOBinding
else:
    import idlelib.IOBinding
    from idlelib.IOBinding import IOBinding


class MyIOBinding(IOBinding):
    def __init__(self, editwin):
        # F5保存时，调用idlelib.runscript.getfilename()，设置自动保存时进入self.editwin.io.save(None)进行保存
        self.save = wrap_function(self.save, editwin.before_save, editwin.after_save)
        IOBinding.__init__(self, editwin)


if PY36:
    idlelib.iomenu.IOBinding = MyIOBinding
else:
    idlelib.IOBinding.IOBinding = MyIOBinding


# - EditorWindow ----------------------------------------


# editor.EditorWindow -> pyshell.PyShellEditorWindow
# editor.EditorWindow -> outwin.OutputWindow -> pyshell.PyShell

if PY36:
    import idlelib.editor
    from idlelib.editor import EditorWindow
else:
    import idlelib.EditorWindow
    from idlelib.EditorWindow import EditorWindow


class MyEditorWindow(EditorWindow):
    def __init__(self, *args):
        if ('advance', 'Advance') not in self.menu_specs:
            self.menu_specs.append(('advance', 'Advance'))

        self.before_copy = []
        self.after_save  = []
        self.before_save = []
        self.before_close = []

        # must before text binding, so before `EditorWindow.__init__()`
        self.cut = wrap_function(self.cut, before=self.before_copy)  # same as `copy`
        self.copy = wrap_function(self.copy, before=self.before_copy)
        self._close = wrap_function(self._close, before=self.before_close)  # "<<close-window>>"事件不命中点击窗口关闭事件

        EditorWindow.__init__(self, *args)
        self.text.tag_lower('hit', 'sel') # fix can't highlight text in sys.stdout

        self.amenu = self.menudict['advance']
        self.make_rmenu() # make "self.rmenu"

        self.recent_files_menu['postcommand'] = self.update_recent_files_list  # fix list not refresh when open another IDLE.

        self.text.bind('<ButtonRelease-1>', lambda e: self.text.event_generate('<<set-line-and-column>>')) # fix event can't multi-binding to `ButtonRelease-1`
        self.load_adv_extensions()
        self.text.bind('<F12>', self.test)

    def add_adv_menu(self, label, sub, index='end', sp=False):
        menu = self.menudict['advance']
        if sp and menu.index('end') is not None:
            menu.insert_separator(index)
        if callable(sub):
            menu.insert_command(index, label=label, command=sub)
        else:
            menu.insert_cascade(index, label=label, menu=sub)

    def load_extension(self, name):
        # for PY34 always raise error.
        if name == 'ZzDummy':
            return
        return super().load_extension(name)

    def load_adv_extensions(self):
        for file in EXTENSIONS:
            name, ext = os.path.splitext(os.path.basename(file))
            if ext == '.py' and name not in ('__init__', '__main__'):
                try:
                    self.load_extension(name) # TODO 支持任意位置文件导入
                except Exception as e:
                    print("Fail to load extension 'idlealib.%s':\n  %s" % (name, e), file=sys.stderr)
                    # import traceback
                    # traceback.print_exc()

        menu = self.menudict['advance']
        if menu.type('end') == 'separator':
            menu.delete('end')

    def test(self, e):
        print('editor on test')
        print('mark_names:', self.text.mark_names())
        print('tag_names:', self.text.tag_names())
        print('functions:', ' '.join(v for v in dir(self.text) if 'tag' in v or 'mark' in v))

        print('wrapped functions:')
        for f1 in ('copy', 'save', 'close'):
            for f2 in ('before', 'after'):
                f3 = f2 + '_' + f1
                if hasattr(self, f3):
                    print(f3 + ':', ['%s.%s' % (f.__module__, f.__name__) for f in getattr(self, f3)])


if PY36:
    idlelib.editor.EditorWindow = MyEditorWindow
else:
    idlelib.EditorWindow.EditorWindow = MyEditorWindow


# - Main ----------------------------------------


# must after hot patch
if PY36:
    from idlelib.pyshell import main
else:
    from idlelib.PyShell import main


def run(ext=None, filename=None):
    if ext:
        EXTENSIONS.clear()
        EXTENSIONS.append(ext)
    fix_path()
    if filename:
        sys.argv.append(filename) # Idea from "~\Lib\tkinter\__main__.py"
    main()


def test_editor(script):
    run(script, script)


def test_shell(script):
    run(script)


if __name__ == '__main__':
    run(None, __file__)


'''
TODO 参考历史文件的打开方法，用于拖拽打开和恢复打开文件
def __recent_file_callback(self, file_name):
    def open_recent_file(fn_closure=file_name):
        self.io.open(editFile=fn_closure)
    return open_recent_file

# 打开文件并定位位置（恢复打开）
outwin.py:
self.flist.gotofileline(filename, lineno)
'''

# TODO 当不在instance_dict中将不会刷新（新启动的idle线程）
'''
def update_recent_files_list(self, new_file=None):
    ...
    # for each edit window instance, construct the recent files menu
    for instance in self.top.instance_dict:
        menu = instance.recent_files_menu
'''



'''文本比较'''

# TODO 优先比对已经打开的文件（列表）


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import sys
import tkinter as tk

if sys.version_info > (3, 6):
    from idlelib.config import idleConf
else:
    from idlelib.configHandler import idleConf

diff_path = idleConf.GetUserCfgDir() + '/diff.html'


def ReadFile(file):
    with open(file, encoding='u8') as f:
        return f.read()


def HtmlDiff(s1, s2):
    import difflib
    import webbrowser
    d = difflib.HtmlDiff()
    with open(diff_path, 'w', encoding='u8') as f:
        f.write(d.make_file(s1.split('\n'), s2.split('\n')))
    webbrowser.open(diff_path)


class CompareFile(tk.Menu):
    def __init__(self, parent):
        tk.Menu.__init__(self, parent.menubar, tearoff=0)
        self.text = parent.text
        self.io = parent.io

        parent.add_adv_menu('Compare to File', self.CompareFile, sp=True)
        parent.add_adv_menu('Compare to Clipboard', self.CompareClipboard)

    def CompareFile(self):
        p2 = self.io.askopenfile()
        if p2:
            s1 = self.text.get('1.0', 'end-1c')
            s2 = ReadFile(p2)
            HtmlDiff(s1, s2)

    def CompareClipboard(self):
        s1 = self.text.get('sel.first', 'sel.last')
        s2 = self.text.clipboard_get()
        HtmlDiff(s1, s2)



'''提取说明文件到readme.md'''

import os
import glob
import imp

os.chdir('..')
for file in glob.iglob('*.py'):
    pkg = imp.load_package(file, file) # every package's `name` should not repeat.
    print('- __%s__  '%file.replace('_', '\\_'))
    print(str(pkg.__doc__).strip().replace('\n', '  \n'))
    print()



'''最近关闭列表'''

# TODO 合并到 RecentSaved.py


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os
import csv
import tkinter as tk

import sys
if sys.version_info > (3, 6):
    from idlelib.config import idleConf
else:
    from idlelib.configHandler import idleConf

userdir = idleConf.GetUserCfgDir()
rc_path = os.path.join(userdir, 'recent-saved.lst')
if not os.path.exists(rc_path):
    open(rc_path, 'w').close()


def ReadData():
    with open(rc_path, encoding='u8', newline='') as f:
        data = list(csv.reader(f))
    return data


class RecentClosed(tk.Menu):
    def __init__(self, parent):
        tk.Menu.__init__(self, parent.menubar, tearoff=0, postcommand=self.Update)
        self.io = parent.io
        parent.amenu.insert_cascade(3, label='Recent Edit Files', menu=self)

    def Update(self): # TODO 排除已打开文件
        rf_list = ReadData()
        ulchars = "1234567890ABCDEFGHIJK"
        rf_list = rf_list[0:len(ulchars)]
        self.delete(0, 'end')
        for i, (cur, file_name) in enumerate(rf_list):
            f = self.Callback(file_name)
            self.add_command(label=ulchars[i] + " " + file_name, command=f, underline=0)

    def Callback(self, file):
        def f():
            self.io.open(editFile=file)
        return f




'''可选中函数提示'''

# TODO 显示完整的__doc__内容并自动调整文本框大小，需要修改idlelib.calltip.get_argspec函数，可能没有简洁的实现方法。


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import tkinter as tk
# import idlelib.calltip


# _MAX_LINES = idlelib.calltip._MAX_LINES
# _MAX_COLS = idlelib.calltip._MAX_COLS
#
#
# class MyCalltip(idlelib.calltip.Calltip):
#     def fetch_tip(self, expression, max_lines=_MAX_LINES, max_cols=_MAX_COLS):
#         idlelib.calltip._MAX_LINES = max_lines
#         idlelib.calltip._MAX_COLS = max_cols
#         self.expression = expression
#         return super().fetch_tip(expression)
#
#     def open_calltip(self, evalfuncs):
#         super().open_calltip(evalfuncs)
#
#         calltip = self.active_calltip
#         if not calltip: # not callable before "(".
#             return
#
#         self.top = calltip.tipwindow
#         self.text = calltip.anchor_widget
#         self.label = calltip.label
#         self.s = calltip.text
#         self.first = False
#
#         # Example:
#         # s2 = '\n'.join('%d: '%(ln+1) + line*2 for ln, line in enumerate(self.s.split('\n')))
#         # self.label.config(text=s2) # hack in !!!
#
#         self.text.bind('<FocusOut>', print) # unbind auto hide. TODO 运行完移除，注意别的影响
#
#         # self.label.bind('<Button>', self.OnClick)  # Exp 1: click and insert
#         self.label.bind('<Enter>', self.OnEnter) # Exp 2: enter and change widget to tk.Text
#
#         # Exp 3: show first as tk.Text and reshape to same size
#         # self.top.update()
#         # self.Label2Text(self.s)
#
#     def OnClick(self, e):
#         h = self.label.winfo_height()
#         y = e.y
#         lines = self.s.split('\n')
#         ln = y // (h // len(lines))
#         self.text.insert('insert', lines[ln])
#
#     def OnEnter(self, e):
#         s2 = self.fetch_tip(self.expression, 999, 999)
#         self.Label2Text(s2)
#
#     def Label2Text(self, s):
#         if self.first:
#             return
#
#         self.first = True
#         self.label.forget()
#
#         # TODO 没有双击选中了
#         text2 = tk.Text(self.top, background='#ffffe0', relief='solid', borderwidth=1, font=self.text['font'])
#         text2.insert('1.0', s)
#         text2.pack(expand=1, fill='both')
#
#         h = self.top.winfo_height()
#         w = self.top.winfo_width()
#         self.top.geometry('%dx%d'%(w, h))
#
#
# idlelib.calltip.Calltip = MyCalltip


class CalltipSelect:
    def __init__(self, parent):
        self.text = parent.text
        self.ctip = parent.ctip

        self.ctip._open_calltip = self.ctip.open_calltip
        self.ctip.open_calltip = self.open_calltip # Good: add actions after original functions

    def open_calltip(self, e):
        self.ctip._open_calltip(False)
        self.first = False
        self.text.unbind('<FocusOut>') # unbind auto hide. TODO 运行完移除，注意别的影响

        calltip = self.ctip.active_calltip

        if calltip:
            self.top   = calltip.tipwindow
            self.label = calltip.label
            self.s     = calltip.text

            self.top.update()
            self.Label2Text(self.s)

    def Label2Text(self, s):
        if self.first:
            return
        self.first = True

        self.label.forget()
        text2 = tk.Text(self.top, background='#ffffe0', relief='solid', borderwidth=1, font=self.text['font'])
        text2.insert('1.0', s)
        text2.config(state='disabled')
        text2.pack(expand=1, fill='both')
        # TODO 没有双击选中了

        h = self.top.winfo_height()
        w = self.top.winfo_width()
        self.top.geometry('%dx%d' % (w, h))



'''插入时间戳注释'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import time


class TimeTag:
    def __init__(self, parent):
        self.text = parent.text

        parent.text.bind('<F4>', self.Insert)

    def Insert(self, evt):
        tag = time.strftime(' # %Y-%m-%d %H:%M:%S')
        header = self.text.get('insert linestart', 'insert')
        if not header.strip():
            tag = tag[1:] + '\n' + header
        self.text.insert('insert', tag)



"""清空Shell"""

# TODO 删除一次Squeezer后再次Squeeze会报错（但不影响其他）


if __name__ == '__main__':
    import __init__
    __init__.test_shell(__file__)


class ClearShell:
    def __init__(self, parent):
        if hasattr(parent, 'write'): # is shell?
            self.text = parent.per.bottom # TODO 研究一下作用原理

            parent.rmenu.add_separator()
            parent.rmenu.add_command(label='Clear shell', command=self.Clear)
            parent.add_adv_menu('Clear shell', self.Clear)

    def Clear(self):
        if self.text.tag_ranges('sel'):
            self.text.delete('sel.first', 'sel.last')
            return

        find = False
        for tag in ['stdout', 'stderr']:
            ranges = self.text.tag_ranges(tag)
            for i in range(len(ranges), 0, -2):
                st, ed = ranges[i-2], ranges[i-1]
                self.text.delete(st.string + 'linestart', ed) # `linestart` for delete Squeezed text
                find = True

        if not find:
            self.text.delete('1.0', 'iomark linestart')


# Ref: ~\idlexlib\extensions\ClearWindow.py



import os
import re
import sys


def lc(s):
    ss = s.split('\n')
    return len(ss), len(ss[-1])


def FindKey(key, path='lib/idlelib', easy=False):
    print('\nFindkey:\n  "%s" in "%s":'%(key, path))
    for root, folders, files in os.walk(os.path.join(sys.base_prefix, path)):
        for file in files:
            if file.endswith('.py'):
                with open(os.path.join(root, file), encoding='u8') as f:
                    s = f.read()
                for m in re.finditer(key, s):
                    row, col = lc(s[:m.start()])
                    file2 = os.path.join(root, file)
                    if easy:
                        row, col = lc(s[:m.start()])
                        print('%s: %d.%d, %s' % (file, row, col, m.group().strip()))
                    else:
                        print('-' * 20)
                        print('File "%s", line %d, col %d:' % (file2, row, col))
                        print('  ' + m.group().strip())
    print()


def PrintTags(text):
    print(text.tag_names())
    for name in text.tag_names():
        print(name, text.tag_ranges(name))


if __name__ == '__main__':
    # FindKey('hello......')

    # FindKey('.*Escape.*')
    FindKey('.*ScriptBinding.*')

    # FindKey('.*<<run-module>>.*')
    # FindKey('.*<<run-module>>.*', easy=True)
    #
    # FindKey('.*<<save-window>>.*')
    # FindKey(r'.*\bmenudefs\b.*')

    # FindKey(r'.*\bCopy\b.*')
    # FindKey(r'.*\b\.tcl\b.*', path='lib/tkinter')

    # FindKey(r'.*EditorWindow.*')

    # FindKey(r'.*<<.*>>.*')

    # FindKey(r'FocusIn')
    # FindKey(r'FOCUSIN')

    # FindKey(r".*\bListedToplevel\b.*")
    # FindKey(r".*\bunregister_callback\b.*")
    # FindKey(r".*\bdict\.\b.*")



"""帮助查看器"""


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import io

import sys
if sys.version_info > (3, 6):
    from idlelib.textview import view_text
else:
    from idlelib.textView import view_text


sp = lambda c: eval(c.replace('.',',')) # Good!
jn = lambda x,y: '%i.%i'%(x,y) # Good!

cmd = '''if 1:
    obj = %r
    try:
        help(eval(obj))
    except (NameError, SyntaxError):
        help(obj)'''


class HelpViewer:
    def __init__(self, parent):
        self.text = parent.text
        self.flist = parent.flist

        parent.rmenu.add_separator()
        parent.rmenu.add_command(label='Help Viewer', command=self.view)
        parent.add_adv_menu('Help Viewer', self.view)

    def view(self):
        sel = self.text.get('sel.first', 'sel.last')
        if sel:
            s = self.help(cmd % sel)
            view_text(self.text, 'Help on %r:' % sel, s, modal=False)

    def help(self, cmd):
        # open shell
        shell = self.flist.pyshell or self.flist.open_shell()
        interp = shell.interp

        # fake printer
        f = io.StringIO()
        _write = shell.write
        shell.write = lambda s, tags=(): f.write(s)

        # run command
        interp.runcommand(cmd)

        # restore
        shell.write = _write

        # get result
        f.seek(0)
        return f.read()



'''仅拷贝Shell的代码'''

# TODO 添加顶部菜单


if __name__ == '__main__':
    import __init__
    __init__.test_shell(__file__)


def GetCodes(text):
    if text.tag_ranges('sel'):
        sted = 'sel.first,sel.last'
    elif text.index('restart'):
        if text.compare('insert', '>', 'restart'):
            sted = 'restart,end'
        else:
            sted = '1.0,restart'
    else:
        sted = '1.0,end'
    st, ed = sted.split(',')
    if not text.tag_ranges('sel'):
        text.tag_add('sel', st, ed) # Select the area to see the selection.

    st_range = text.tag_nextrange('stdin', st)
    ed_range = text.tag_prevrange('stdin', ed)
    ranges = [cur.string for cur in text.tag_ranges('stdin')]
    ranges = [tuple(ranges[i:i + 2]) for i in range(0, len(ranges), 2)]
    if st_range and ed_range:
        ranges = ranges[ranges.index(st_range):ranges.index(ed_range) + 1]
        codes = [text.get(*r) for r in ranges]
        s = ''.join(codes)
        return s.strip()
    return ''


class SmartCopy:
    def __init__(self, parent):
        if not hasattr(parent, 'write'): # is not shell?
            return

        self.text = parent.text
        self.io = parent.io
        self.flist = parent.flist

        n = 5
        parent.rmenu.insert_separator(n)
        parent.rmenu.insert_command(n, label='Export to File', command=self.New)
        parent.rmenu.insert_command(n, label='Copy Code', command=self.Copy)

    def Copy(self):
        self.text.clipboard_clear()
        self.text.clipboard_append(GetCodes(self.text))

    def New(self): # See: idlelib.editor.EditorWindow.new_callback
        dirname, basename = self.io.defaultfilename()
        editwin = self.flist.new(dirname)
        editwin.text.insert('1.0', GetCodes(self.text))



"""带参数运行"""


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


from tkinter.simpledialog import askstring


class RunArgs:
    def __init__(self, parent):
        if hasattr(parent, 'write'): # is shell?
            return

        self.text = parent.text
        self.root = parent.root
        self.flist = parent.flist

        parent.add_adv_menu('Run with args', self.Run, sp=True)

    def Run(self):
        args = askstring('Run with args', 'args:', parent=self.root)
        if args is not None:
            cmd = "__import__('sys').argv.extend(%r.split())" % args
            shell = self.flist.open_shell()
            interp = shell.interp
            interp.runcommand(cmd)
            interp._runcode = interp.runcode
            interp.runcode = lambda evt: [interp.runcommand(cmd), interp._runcode(evt)]
            self.text.event_generate('<<run-module>>')


if __name__ == '__main__': # for test
    import sys
    print(sys.argv)



'''重载文件'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os
from tkinter.messagebox import askyesno


def mtime(file):
    return os.path.getmtime(file) if file else 0


class AutoReload:
    def __init__(self, parent):
        self.text = parent.text
        self.io = parent.io
        self.set_saved = parent.set_saved

        self.mt = mtime(self.io.filename)
        parent.text_frame.bind('<FocusIn>', self.OnFocusIn)
        parent.after_save.append(self.Refresh)
        parent.add_adv_menu('Reload', self.ReloadFile)

    def Refresh(self):
        self.mt = mtime(self.io.filename)

    def ReloadFile(self): # TODO 重载记忆位置
        self.io.loadfile(self.io.filename)  # TODO rename也可以用这个方法

    def OnFocusIn(self, e):
        mt = mtime(self.io.filename)
        if self.mt != mt:
            self.mt = mt  # 激活窗口最多只提示一次，下一次提示在本地文件再次发生修改
            if askyesno('Reload', '"%s"\n\nThis script has been modified by another program.\nDo you want to reload it?' % self.io.filename, parent=self.text):
                self.ReloadFile()
            else:
                self.set_saved(False) # 设置为未保存状态



import os
import shutil
import setuptools

keys = ['readme', 'license', 'todo', 'version']
os.makedirs('idlealib/docs', exist_ok=True)
for file in os.listdir():
    if any(file.lower().startswith(k) for k in keys):
        shutil.copy(file, 'idlealib/docs/' + file.lower())

with open('VERSION.txt') as f:
    version = f.read().strip()

with open('README.md', encoding='u8') as f:
    long_description = f.read().replace('[x]', chr(9989)).replace('[ ]', chr(10062)) # ✅❎

setuptools.setup(
    name='idlea',
    version=version,
    author='Lishixian(znsoooo)',
    author_email='lsx7@sina.com',
    description='IDLE-Advance',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/znsoooo/IDLE-Advance',
    project_urls={
        'Bug Tracker': 'https://github.com/znsoooo/IDLE-Advance/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Framework :: IDLE',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Text Editors :: Integrated Development Environments (IDE)',
    ],
    packages=['idlealib', 'idlealib.docs', 'idlealib.nouse', 'idlealib.scripts'],
    python_requires='>=3.4',

    # install_requires=['windnd'], # for drag-open file feature
    # extras_requires={'windnd': ['windnd']},
    license='MIT License',
    entry_points={'console_scripts': ['idlea=idlealib:run']},
    package_data={'': ['*.*']},
    keywords='IDLE-Advance IDLE IDLEA IDLEX extension idlelib idlexlib idlealib',
)



'''配置右键菜单、桌面快捷方式、开始菜单、右键新建的界面菜单'''

# TODO 自动匹配pythonw.exe

import os
import sys
import time
import winreg
import tkinter as tk
import tkinter.ttk as ttk
from tkinter.messagebox import showinfo

options = ('All Positions Below', 'Lnk: Desktop Shortcut', 'Lnk: Start Up Menu', 'Reg: Right Click Menu', 'Reg: Create New File')


# ---------------------------------------------------------------------------
# make link
# ---------------------------------------------------------------------------


def MakeLink(exe, folder, file, cmd):
    try:
        import winshell
        root = os.path.dirname(exe)
        path = os.path.join(folder, file)
        winshell.CreateShortcut(path, exe, cmd, root, (root + r'\Lib\idlelib\Icons\idle.ico', 0))
    except ImportError:
        os.popen('explorer "%s"' % folder)
        top.clipboard_clear()
        top.clipboard_append('"%s" %s' % (exe, cmd))
        time.sleep(0.5)
        showinfo('Info', 'Copy `cmd` and create shortcut by right click in opened explorer')


# ---------------------------------------------------------------------------
# edit regedit
# ---------------------------------------------------------------------------


def GetDir(name, local=False): # Name: 'Desktop', 'SendTo', 'Programs', 'Startup'
    path = r'Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders'
    if local:
        name = 'Common ' + name
        key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, path)
    else:
        key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, path)
    return winreg.QueryValueEx(key, name)[0]


def DeleteRecu(key):
    for i in range(winreg.QueryInfoKey(key)[0]):
        name = winreg.EnumKey(key, i)
        DeleteRecu(winreg.OpenKey(key, name))
        winreg.DeleteKey(key, name)


def DeleteKey(key, name=''):
    root, p = key.split('\\', 1)
    key2 = winreg.OpenKey(getattr(winreg, root), p)
    if name:
        winreg.DeleteValue(key2, name)
    else:
        DeleteRecu(key2)
        winreg.DeleteKey(getattr(winreg, root), p)


def SetKey(key, name='', val=''):
    root, p = key.split('\\', 1)
    key2 = winreg.CreateKey(getattr(winreg, root), p)
    winreg.SetValueEx(key2, name, 0, winreg.REG_SZ, val)


# ---------------------------------------------------------------------------
# GUI
# ---------------------------------------------------------------------------


def Loop(fun):
    for name in options[1:]:
        v0.set(name)
        Switch()
        fun()
    v0.set(options[0])
    Switch()


def Add():
    id = options.index(v0.get())
    s1, s2, s3, s4 = v1.get(), v2.get(), v3.get(), v4.get()
    if id == 0:
        Loop(Add)
        return
    try:
        if id in [1, 2]: # lnk
            MakeLink(s1, s2, s3, s4)
        elif id in [3, 4]: # reg
            SetKey(s2, s3, s4)
    except Exception as e:
        showinfo('Error', str(e))
    # else:
    #     showinfo('Info', 'Success!')


def Remove():
    id = options.index(v0.get())
    s2, s3, s4 = v2.get(), v3.get(), v4.get()
    if id == 0:
        Loop(Remove)
        return
    try:
        if id in [1, 2]: # lnk
            p = os.path.join(s2, s3)
            if os.path.isfile(p):
                os.remove(p)
        elif id in [3, 4]: # reg
            DeleteKey(s2, s3)
    except Exception as e:
        showinfo('Error', str(e))
    # else:
    #     showinfo('Info', 'Success!')


def Switch(*_):
    id = options.index(v0.get())
    exe = v1.get()
    src = os.path.abspath('../__main__.py')
    s2 = s3 = s4 = ''
    if id == 1:
        s2 = GetDir('Desktop')
        s3 = 'IDLE-Advance.lnk'
        s4 = '"%s"' % src
    elif id == 2:
        s2 = GetDir('Programs')
        s3 = 'Python IDLE-Advance.lnk'
        s4 = '"%s"' % src
    elif id == 3:
        s2 = r'HKEY_CURRENT_USER\Software\Classes\Python.File\Shell\Edit with IDLE-Adv\command'
        s4 = '"{}" "{}" "%L" %*'.format(exe, src)
    elif id == 4:
        s2 = r'HKEY_CURRENT_USER\Software\Classes\.py\ShellNew'
        s3 = 'FileName'
    v2.set(s2)
    v3.set(s3)
    v4.set(s4)


def Center(self):
    self.withdraw() # withdraw/deiconify 阻止页面闪烁
    self.update_idletasks()
    x = (self.winfo_screenwidth() - self.winfo_reqwidth()) / 2
    y = (self.winfo_screenheight() - self.winfo_reqheight()) / 2
    self.geometry('+%d+%d' % (x, y))
    self.deiconify()


if __name__ == '__main__':
    top = tk.Tk()
    top.title('IDLE-Advance Context Helper')
    top.resizable(1, 0)
    top.columnconfigure(1, weight=1) # for `grid` can expand

    v0 = tk.StringVar(value='All')
    v1 = tk.StringVar(value=sys.executable)
    v2 = tk.StringVar(value='')
    v3 = tk.StringVar(value='')
    v4 = tk.StringVar(value='')

    v1.trace_add('write', Switch)

    text = tk.Label(top, text=' Add to: ')
    menu = ttk.OptionMenu(top, v0, options[0], *options, command=Switch)
    menu.config(width=20)
    btn1 = ttk.Button(top, text='Add',    command=Add)
    btn2 = ttk.Button(top, text='Remove', command=Remove)

    text.grid(row=0, column=0, sticky='e')
    menu.grid(row=0, column=1, sticky='w')
    btn1.grid(row=0, column=2)
    btn2.grid(row=0, column=3)

    for r, (name, var) in enumerate(zip(['exe', 'path', 'name', 'cmd'], [v1, v2, v3, v4])):
        tk.Label(top, text=' %s: ' % name).grid(row=r+1, column=0, sticky='e')
        tk.Entry(top, textvariable=var)   .grid(row=r+1, column=1, sticky='we', columnspan=3)

    Center(top)

    top.mainloop()



'''搜索替换工具条'''

# TEST TEXT
# A quick brown fox
# JuMp oVeR THE lazy dog
# 1377
# [a-z]+

# TODO 处理好和纵向滚动条的相对位置问题


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import re
import tkinter as tk

import sys
PY37 = sys.version_info > (3, 7)
PY36 = sys.version_info > (3, 6)
if PY36:
    from idlelib import searchengine
    from idlelib.replace import ReplaceDialog
    from idlelib.search import _setup
else:
    import idlelib.SearchEngine as searchengine
    from idlelib.ReplaceDialog import ReplaceDialog
    from idlelib.SearchDialog import _setup


jn = lambda x,y: '%i.%i'%(x,y) # Good!
sp = lambda c: eval(c.replace('.',',')) # Good!
lc = lambda s: jn(s.count('\n')+1, len(s)-s.rfind('\n')-1) # Good!


class ReplaceBar(tk.Frame):
    def __init__(self, parent):
        tk.Frame.__init__(self, parent.text_frame)

        self.root = parent.root
        self.text = parent.text
        self.text.replace_bar = self # hook

        self.engine = engine = searchengine.get(self.root)
        self.replace = replace = ReplaceDialog(self.root, engine)
        self.show = False

        # hot fix for `dialog.open(text)` effect called in `idlelib.replace.replace`
        replace.text = self.text
        replace.ok = 1
        replace.bell = parent.top.bell

        # class `SearchDialog` is separate with `ReplaceDialog`
        search = _setup(self.text)
        search.open = self.Open

        engine.patvar .trace('w', self.Update)
        engine.revar  .trace('w', self.Update)
        engine.casevar.trace('w', self.Update)
        engine.wordvar.trace('w', self.Update)

        t1 = tk.Entry(self, width=8, textvariable=engine.patvar)
        t2 = tk.Entry(self, width=8, textvariable=replace.replvar)
        tk.Label(self, text='Find:').pack(side='left')
        t1.pack(side='left', fill='x', expand=True)
        tk.Label(self, text='Repl:').pack(side='left')
        t2.pack(side='left', fill='x', expand=True)
        self.t1 = t1

        self.tip = tk.Label(self, text=' Match: 0')
        self.tip.pack(side='left')

        tk.Checkbutton(self, text='Re', variable=engine.revar)  .pack(side='left')
        tk.Checkbutton(self, text='Cc', variable=engine.casevar).pack(side='left')
        tk.Checkbutton(self, text='Wd', variable=engine.wordvar).pack(side='left')

        tk.Button(self, relief='groove', text='<<', command=lambda: self.Find(0)).pack(side='left')
        tk.Button(self, relief='groove', text='>>', command=lambda: self.Find(1)).pack(side='left')
        tk.Button(self, relief='groove', text='Replace', command=self.Replace).pack(side='left')
        tk.Button(self, relief='groove', text='Replace All', command=self.ReplaceAll).pack(side='left')

        self.text.bind("<<replace>>", self.OnReplace)

        t1.bind('<FocusIn>', self.Update)
        self.text.bind('<Key-Escape>', self.Hide)
        t1.bind('<Escape>', self.Hide)
        t2.bind('<Escape>', self.Hide)

    def OnReplace(self, evt):
        self.text.event_generate('<<find>>')
        return 'break'

    def Open(self, text, string=None):
        # remove selection to avoid replacing only itself in `replace_all` function
        text.tag_remove('sel', '1.0', 'end-1c')

        bar = text.replace_bar
        if PY37:
            bar.grid(row=3, column=1, sticky='nsew')
        else:
            bar.pack(fill='x', side='bottom')

        if string:
            bar.engine.setpat(string)
        bar.t1.focus()
        bar.t1.select_range(0, 'end')  # don't use `t1.select_to('end')`. 选中区域为空时，输入字符后无法删除（移动光标后可以删除）
        bar.t1.icursor('end')
        bar.show = True
        bar.Update()

    def Hide(self, evt):
        if PY37:
            self.grid_forget()
        else:
            self.forget()
        self.text.tag_remove('hit', '1.0', 'end')
        self.text.focus()
        self.show = False

    def Update(self, *args):
        if not self.show:
            return

        self.text.tag_remove('hit', '1.0', 'end')
        self.tip.config(text=' Match: 0')

        # banning show error report temporarily
        _report_error = self.engine.report_error
        self.engine.report_error = lambda *v: None
        prog = self.engine.getprog()
        self.engine.report_error = _report_error

        if not prog:
            return

        s = self.text.get('1.0', 'end-1c')
        insert = len(self.text.get('1.0', 'insert'))
        matchs = [m.span() for m in re.finditer(prog.pattern, s, prog.flags | re.M)] # fix matches the end of lines
        self.tip.config(text=' Match: %d' % len(matchs))
        if len(matchs) < 1000:
            for n, (p1, p2) in enumerate(matchs):
                if p1 <= insert < p2:
                    self.tip.config(text=' Match: %d/%d' % (n + 1, len(matchs)))
                self.text.tag_add('hit', lc(s[:p1]), lc(s[:p2])) # 如果使用`1.0+nc`字符偏移会命中Squeezer导致错位

    def Find(self, forward):
        self.engine.wrapvar.set(True)
        self.engine.backvar.set(not forward)
        self.replace.find_it()
        self.Update()

    def Replace(self):
        self.replace.default_command()
        self.text.after(5, self.Update) # TODO unknown reason

    def ReplaceAll(self):
        # Ref: idlelib.replace.replace_all
        text = self.text

        prog = self.engine.getprog()
        if not prog: return

        res = self.engine.search_text(text, prog)
        if not res: return

        text.undo_block_start()

        sel_first = text.index('sel.first')
        line, col = sp(text.index('sel.first') or '1.0')
        line3, col3 = sp(text.index('sel.last') or text.index('end-1c'))
        text.tag_remove('sel', '1.0', 'end')
        text.tag_remove('hit', '1.0', 'end')

        repl = self.replace.replvar.get()
        while 1:
            res = self.engine.search_forward(text, prog, line, col, 0)
            if not res: break

            line, m = res
            new = self.replace._replace_expand(m, repl)
            if new is None: break

            col1, col2 = m.span()
            col = col1 + len(new)
            if (line, col2) > (line3, col3):
                break
            if line == line3:
                col3 += col - col2

            first = jn(line, col1)
            last  = jn(line, col2)
            text.delete(first, last)
            text.insert(first, new)

        if sel_first:
            text.tag_add('sel', sel_first, jn(line3, col3))

        text.undo_block_stop()

        self.text.after(5, self.Update) # TODO unknown reason



'''窗口操作'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os
import tkinter as tk


class WindowManager(tk.Menu):
    def __init__(self, parent):
        tk.Menu.__init__(self, parent.menubar, tearoff=0)
        self.parent = parent
        self.text = parent.text
        self.flist = parent.flist
        self.MakeMenu()
        self.Binding()
        parent.amenu.insert_cascade(3, label='Window Manager', menu=self)

    def OnRestore(self, *args): # TODO 恢复窗口时恢复记忆位置
        with open(self.parent.recent_files_path, encoding='u8') as f:
            rf_list = f.read().splitlines()
        dict = self.flist.dict
        for file in rf_list:
            if os.path.normcase(file) not in dict:
                return self.parent.io.open(editFile=file)

    def OnNext(self, n):
        windows = list(self.flist.inversedict)
        index = windows.index(self.parent)
        window = windows[(index + n) % len(windows)]
        window.top.wakeup()  # See at: idlelib.window.WindowList.add_windows_to_menu
        return 'break' # 避免窗口左右移动

    def OnNew(self, *args):
        self.text.event_generate('<<open-new-window>>')

    def OnClose(self, *args):
        self.OnNext(-1)
        self.text.event_generate('<<close-window>>')

    def OnCloseAll(self, *args):
        self.text.event_generate('<<close-all-windows>>')

    def MakeMenu(self):
        self.add_command(label='New Window',        command=self.OnNew)
        self.add_command(label='Restore Window',    command=self.OnRestore)
        self.add_separator()
        self.add_command(label='Prev Window',       command=lambda: self.OnNext(-1))
        self.add_command(label='Next Window',       command=lambda: self.OnNext(1))
        self.add_separator()
        self.add_command(label='Close Window',      command=self.OnClose)
        self.add_command(label='Close All Windows', command=self.OnCloseAll)

    def Binding(self):
        text = self.text
        text.bind('<Control-t>',       self.OnNew)
        text.bind('<Control-Shift-T>', self.OnRestore)
        text.bind('<Control-Prior>',   lambda e: self.OnNext(-1))
        text.bind('<Control-Next>',    lambda e: self.OnNext(1))
        text.bind('<Control-w>',       self.OnClose)
        text.bind('<Control-Shift-W>', self.OnCloseAll)



"""平移多行"""


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


class ShiftLines:
    def __init__(self, parent):
        self.text = parent.text
        self.text.bind('<Alt-Up>', self.callback)
        self.text.bind('<Alt-Down>', self.callback)

    def callback(self, evt):
        text = self.text
        if not text.tag_ranges('sel'):
            text.tag_add('sel', 'insert linestart', 'insert+1l linestart')
        else:
            text.tag_add('sel', 'sel.first linestart', 'sel.last-1c+1l linestart')
        text.mark_set('insert', 'sel.first')
        text.undo_block_start()
        if evt.keysym == 'Up':
            s = text.get('sel.first-1l', 'sel.first')
            text.delete('sel.first-1l', 'sel.first')
            text.insert('sel.last', s)
        else:
            s = text.get('sel.last', 'sel.last+1l')
            text.delete('sel.last', 'sel.last+1l')
            text.insert('sel.first', s)
        text.undo_block_stop()
        return 'break'



'''匹配输入和删除成对括号'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


LEFT = '([{\'"`'
RIGHT = ')]}'
pair = lambda c: ')]}\'"`'['([{\'"`'.index(c)]


def GetSpan(text):
    sel = text.tag_ranges('sel')
    left = 'sel.first' if sel else 'insert'
    right = 'sel.last' if sel else 'insert'
    return left, right


class SmartPairing:
    def __init__(self, parent):
        self.text = parent.text
        for c in LEFT + RIGHT:
            self.text.bind('<%s>' % c, self.PairAdd)  # '<KeyRelease-%s>'%c
        self.text.bind('<BackSpace>', self.OnBackSpace)

    def PairAdd(self, evt):
        c = evt.char
        text = self.text

        if c in LEFT: # 输入左括号时输入匹配右括号，允许含有选区
            text.undo_block_start()

            if text.tag_ranges('sel'):
                c1 = text.get('sel.first')
                if c1 in LEFT and c1 != c and ((c1 in '([{') ^ (c in '\'"`')):
                    self.PairDelete()

            sel = text.tag_ranges('sel')
            left = 'sel.first' if sel else 'insert'
            right = 'sel.last' if sel else 'insert'

            text.insert(left, c)
            text.mark_gravity('insert', 'left')
            text.insert(right, pair(c))
            text.mark_gravity('insert', 'right')
            text.undo_block_stop()
            return 'break'

        elif c in RIGHT: # 输入右括号时移动光标但不键入
            if c == text.get('insert'):
                text.mark_set('insert', 'insert+1c')
                return 'break'

    def PairDelete(self): # 删除左括号时同时删除右括号
        text = self.text

        sel = text.tag_ranges('sel')
        left = 'sel.first' if sel else 'insert-1c'
        right = 'sel.last-1c' if sel else 'insert'

        c1 = text.get(left)
        c2 = text.get(right)

        if c1 in LEFT and c2 == pair(c1):
            text.undo_block_start()
            text.delete(right)
            text.delete(left)
            text.undo_block_stop()
            return True

    def OnBackSpace(self, evt):
        if not self.PairDelete():
            self.text.event_generate('<<smart-backspace>>')
        return 'break'



"""快捷反缩进"""


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


class ShiftTab:
    def __init__(self, parent):
        self.text = parent.text
        self.text.bind('<Shift-Tab>', self.callback)

    def callback(self, evt):
        self.text.event_generate('<<dedent-region>>')
        return 'break'
